var __values = (this && this.__values) || function (o) {
    var m = typeof Symbol === "function" && o[Symbol.iterator], i = 0;
    if (m) return m.call(o);
    return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
};
import { Subject } from 'rxjs';
import { extractHostnameAndPort, generateId, generateKey, isBrowser, isOnline, isURL, isVisible, log, validateKey, } from './misc/util';
import { ChannelBuilder } from './service/channelBuilder/ChannelBuilder';
import { FullMesh } from './service/topology/FullMesh';
import { TopologyEnum, TopologyState } from './service/topology/Topology';
import { UserMessage } from './service/UserMessage';
import { Signaling, SignalingState } from './Signaling';
import { WebChannelState } from './WebChannelState';
import { WebSocketBuilder } from './WebSocketBuilder';
export var webChannelDefaultOptions = {
    topology: TopologyEnum.FULL_MESH,
    signalingServer: 'wss://signaling.netflux.coedit.re',
    rtcConfiguration: {
        iceServers: [{ urls: 'stun:stun3.l.google.com:19302' }],
    },
    autoRejoin: true,
};
var REJOIN_TIMEOUT = 3000;
var WebChannel = (function () {
    function WebChannel(options) {
        this.STREAM_ID = 2;
        var fullOptions = Object.assign({}, webChannelDefaultOptions, options);
        this.streamSubject = new Subject();
        this.idSubject = new Subject();
        this.topologyEnum = fullOptions.topology;
        this.autoRejoin = fullOptions.autoRejoin;
        this.rtcConfiguration = fullOptions.rtcConfiguration;
        this.members = [];
        this._id = 0;
        this.key = '';
        this.myId = 0;
        this.state = WebChannelState.LEFT;
        this.rejoinEnabled = false;
        this.rejoinTimer = undefined;
        this.topology = {};
        this._onAlone = function () { };
        this.onMemberJoin = function none() { };
        this.onMemberLeave = function none() { };
        this.onMessage = function none() { };
        this.onStateChange = function none() { };
        this.onSignalingStateChange = function none() { };
        this.userMsg = new UserMessage();
        this.signaling = new Signaling(this, fullOptions.signalingServer);
        this.subscribeToSignalingState();
        this.webSocketBuilder = new WebSocketBuilder(this);
        this.channelBuilder = new ChannelBuilder(this);
        this.setTopology(fullOptions.topology);
        if (isBrowser) {
            this.subscribeToBrowserEvents();
        }
    }
    Object.defineProperty(WebChannel.prototype, "onIdChange", {
        get: function () {
            return this.idSubject.asObservable();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(WebChannel.prototype, "id", {
        get: function () {
            return this._id;
        },
        set: function (value) {
            this._id = value;
            this.idSubject.next(value);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(WebChannel.prototype, "messageFromStream", {
        get: function () {
            return this.streamSubject.asObservable();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(WebChannel.prototype, "onAlone", {
        set: function (handler) {
            this._onAlone = handler;
        },
        enumerable: true,
        configurable: true
    });
    WebChannel.prototype.sendOverStream = function (msg) {
        this.topology.sendTo(msg);
    };
    WebChannel.prototype.join = function (key) {
        if (key === void 0) { key = generateKey(); }
        validateKey(key);
        if (this.state === WebChannelState.LEFT) {
            this.startJoin(key);
        }
    };
    WebChannel.prototype.invite = function (url) {
        var e_1, _a;
        if (isURL(url)) {
            var hostnamePort = extractHostnameAndPort(url);
            try {
                for (var _b = __values(this.topology.neighbors), _c = _b.next(); !_c.done; _c = _b.next()) {
                    var ch = _c.value;
                    if (hostnamePort === extractHostnameAndPort(ch.url)) {
                        return;
                    }
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
                }
                finally { if (e_1) throw e_1.error; }
            }
            this.webSocketBuilder
                .connectToInvite(url)
                .catch(function (err) { return log.webgroup("Failed to invite the bot " + url + ": " + err.message); });
        }
        else {
            throw new Error("Failed to invite a bot: " + url + " is not a valid URL");
        }
    };
    WebChannel.prototype.leave = function () {
        if (this.state !== WebChannelState.LEFT) {
            this.key = '';
            this.internalLeave();
        }
    };
    WebChannel.prototype.send = function (data) {
        var e_2, _a;
        if (this.members.length !== 1) {
            try {
                for (var _b = __values(this.userMsg.encodeUserMessage(data)), _c = _b.next(); !_c.done; _c = _b.next()) {
                    var chunk = _c.value;
                    this.topology.send({
                        senderId: this.myId,
                        recipientId: 0,
                        serviceId: UserMessage.SERVICE_ID,
                        content: chunk,
                    });
                }
            }
            catch (e_2_1) { e_2 = { error: e_2_1 }; }
            finally {
                try {
                    if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
                }
                finally { if (e_2) throw e_2.error; }
            }
        }
    };
    WebChannel.prototype.sendTo = function (id, data) {
        var e_3, _a;
        if (this.members.length !== 1) {
            try {
                for (var _b = __values(this.userMsg.encodeUserMessage(data)), _c = _b.next(); !_c.done; _c = _b.next()) {
                    var chunk = _c.value;
                    this.topology.sendTo({
                        senderId: this.myId,
                        recipientId: id,
                        serviceId: UserMessage.SERVICE_ID,
                        content: chunk,
                    });
                }
            }
            catch (e_3_1) { e_3 = { error: e_3_1 }; }
            finally {
                try {
                    if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
                }
                finally { if (e_3) throw e_3.error; }
            }
        }
    };
    WebChannel.prototype.onMemberJoinProxy = function (id) {
        if (!this.members.includes(id)) {
            this.members[this.members.length] = id;
            this.onMemberJoin(id);
        }
    };
    WebChannel.prototype.onAdjacentMembersLeaveProxy = function (ids) {
        if (this.onMemberLeaveProxy(ids)) {
            if (this.members.length === 1) {
                this._onAlone();
                this.topology.leave();
            }
            else if (this.signaling.state === SignalingState.CHECKED &&
                this.topology.state === TopologyState.CONSTRUCTED) {
                this.signaling.check();
            }
        }
    };
    WebChannel.prototype.onDistantMembersLeaveProxy = function (ids) {
        this.onMemberLeaveProxy(ids);
    };
    WebChannel.prototype.init = function (key, id) {
        if (id === void 0) { id = generateId(); }
        log.webgroup('INIT');
        this.id = id;
        this.myId = generateId();
        this.members = [this.myId];
        this.key = key;
        this.rejoinEnabled = this.autoRejoin;
        if (this.rejoinTimer) {
            clearTimeout(this.rejoinTimer);
            this.rejoinTimer = undefined;
        }
        this.setState(WebChannelState.JOINING);
    };
    WebChannel.prototype.clean = function () {
        log.webgroup('CLEAN');
        if (this.rejoinTimer) {
            clearTimeout(this.rejoinTimer);
            this.rejoinTimer = undefined;
        }
        this.members = [];
        this.id = 0;
        this.myId = 0;
    };
    WebChannel.prototype.setState = function (state) {
        if (this.state !== state) {
            log.webGroupState(WebChannelState[state], this.myId);
            this.state = state;
            this.onStateChange(state);
        }
    };
    WebChannel.prototype.setTopology = function (topologyEnum) {
        var _this = this;
        this.topologyEnum = topologyEnum;
        this.topology = new FullMesh(this);
        this.topology.onState.subscribe(function (state) {
            log.webgroup('Topology state: ', TopologyState[state]);
            switch (state) {
                case TopologyState.CONSTRUCTING:
                    _this.setState(WebChannelState.JOINING);
                    break;
                case TopologyState.CONSTRUCTED:
                    if (_this.signaling.state === SignalingState.OPEN) {
                        _this.signaling.check();
                    }
                    else if (_this.signaling.state === SignalingState.CHECKED) {
                        _this.setState(WebChannelState.JOINED);
                        _this.signaling.check();
                    }
                    else if (_this.signaling.state === SignalingState.CLOSED) {
                        _this.signaling.connect(_this.key);
                    }
                    break;
                case TopologyState.IDLE:
                    _this.channelBuilder.clean();
                    _this.userMsg.clean();
                    switch (_this.signaling.state) {
                        case SignalingState.CLOSED:
                            if (_this.rejoinEnabled) {
                                _this.rejoin();
                            }
                            else {
                                _this.internalLeave();
                            }
                            break;
                        case SignalingState.CONNECTING:
                            _this.setState(WebChannelState.JOINING);
                            break;
                        case SignalingState.OPEN:
                            _this.setState(WebChannelState.JOINING);
                            _this.signaling.check();
                            break;
                        case SignalingState.CHECKING:
                            _this.setState(WebChannelState.JOINING);
                            break;
                        case SignalingState.CHECKED:
                            _this.signaling.check();
                            break;
                    }
                    break;
            }
        });
    };
    WebChannel.prototype.subscribeToSignalingState = function () {
        var _this = this;
        this.signaling.onState.subscribe(function (state) {
            log.signalingState(SignalingState[state], _this.myId);
            _this.onSignalingStateChange(state);
            switch (state) {
                case SignalingState.CLOSED:
                    if (_this.topology.state === TopologyState.IDLE) {
                        if (_this.rejoinEnabled) {
                            _this.rejoin();
                        }
                        else {
                            _this.internalLeave();
                        }
                    }
                    else if (_this.topology.state === TopologyState.CONSTRUCTED) {
                        if (_this.members.length === 1) {
                            _this.topology.leave();
                        }
                        else if (_this.rejoinEnabled && !_this.rejoinTimer) {
                            _this.reconnectToSignaling();
                        }
                    }
                    break;
                case SignalingState.OPEN:
                    if (_this.topology.state !== TopologyState.CONSTRUCTING) {
                        _this.signaling.check();
                    }
                    break;
                case SignalingState.CHECKED:
                    if (_this.topology.state === TopologyState.IDLE) {
                        if (_this.signaling.connected) {
                            _this.topology.setJoinedState();
                        }
                    }
                    else if (_this.topology.state === TopologyState.CONSTRUCTED) {
                        _this.setState(WebChannelState.JOINED);
                    }
                    break;
            }
        });
    };
    WebChannel.prototype.subscribeToBrowserEvents = function () {
        var _this = this;
        window.addEventListener('online', function () { return _this.onBrowserBack(); });
        window.addEventListener('visibilitychange', function () { return _this.onBrowserBack(); });
        window.addEventListener('beforeunload', function () { return _this.leave(); });
    };
    WebChannel.prototype.startJoin = function (key) {
        if (key === void 0) { key = this.key; }
        log.webgroup('start join...');
        this.init(key);
        this.signaling.connect(key);
    };
    WebChannel.prototype.rejoin = function () {
        var _this = this;
        this.setState(WebChannelState.JOINING);
        if (!isVisible() || !isOnline()) {
            this.internalLeave();
        }
        else {
            this.clean();
            log.webgroup("rejoin in " + REJOIN_TIMEOUT + "ms");
            this.rejoinTimer = setTimeout(function () {
                if (_this.signaling.state === SignalingState.CLOSED &&
                    isVisible() &&
                    isOnline() &&
                    _this.rejoinEnabled) {
                    _this.startJoin();
                }
                else {
                    log.webgroup('abandon rejoin because: ', {
                        isVisible: isVisible(),
                        isOnline: isOnline(),
                        signalingState: SignalingState[_this.signaling.state],
                        rejoinEnabled: _this.rejoinEnabled,
                    });
                    _this.internalLeave();
                }
                _this.rejoinTimer = undefined;
            }, REJOIN_TIMEOUT);
        }
    };
    WebChannel.prototype.reconnectToSignaling = function () {
        var _this = this;
        if (this.members.length === 1 && (!isVisible() || !isOnline())) {
            this.internalLeave();
        }
        else {
            log.webgroup("reconnect to Signaling server in " + REJOIN_TIMEOUT + "ms");
            this.setState(WebChannelState.JOINING);
            this.rejoinTimer = setTimeout(function () {
                if (_this.signaling.state === SignalingState.CLOSED) {
                    if (isVisible() && isOnline()) {
                        if (_this.rejoinEnabled) {
                            _this.signaling.connect(_this.key);
                        }
                    }
                    else if (_this.members.length === 1) {
                        _this.internalLeave();
                    }
                }
                _this.rejoinTimer = undefined;
            }, REJOIN_TIMEOUT);
        }
    };
    WebChannel.prototype.onBrowserBack = function () {
        if (isVisible() && isOnline()) {
            log.webgroup('onBrowserBack', { isVisible: isVisible(), isOnline: isOnline() });
            this.rejoinEnabled = this.autoRejoin;
            if (this.rejoinEnabled) {
                if (this.state === WebChannelState.LEFT) {
                    this.startJoin();
                }
                else if (this.signaling.state === SignalingState.CLOSED) {
                    this.signaling.connect(this.key);
                }
            }
        }
    };
    WebChannel.prototype.onMemberLeaveProxy = function (ids) {
        var _this = this;
        var atLeastOneLeft = false;
        ids.forEach(function (id) {
            if (_this.members.includes(id)) {
                _this.members.splice(_this.members.indexOf(id), 1);
                atLeastOneLeft = true;
                _this.onMemberLeave(id);
            }
        });
        return atLeastOneLeft;
    };
    WebChannel.prototype.internalLeave = function () {
        log.webgroup('internal leave');
        this.rejoinEnabled = false;
        this.signaling.close();
        this.topology.leave();
        this.clean();
        this.setState(WebChannelState.LEFT);
    };
    return WebChannel;
}());
export { WebChannel };
//# sourceMappingURL=WebChannel.js.map