import { ConnectionError } from './ConnectionError';
var ConnectionsInProgress = (function () {
    function ConnectionsInProgress(wcStreamId) {
        this.wcStreamId = wcStreamId;
        this.wcStream = new Map();
        this.sigStream = new Map();
    }
    ConnectionsInProgress.prototype.create = function (streamId, id, connectionTimeout, responseTimeout, onConnectionTimeoutCallback) {
        var connections = this.getByStreamId(streamId);
        var connection = {};
        connection.promise = new Promise(function (resolveResponse, rejectResponse) {
            var responseTimer = setTimeout(function () { return connection.reject(new Error(ConnectionError.RESPONSE_TIMEOUT)); }, responseTimeout);
            connection.resolve = function () {
                clearTimeout(responseTimer);
                resolveResponse();
                connection.promise = new Promise(function (resolveConnection, rejectConnection) {
                    var connectionTimer = setTimeout(function () {
                        onConnectionTimeoutCallback();
                        connection.reject(new Error(ConnectionError.CONNECTION_TIMEOUT));
                    }, connectionTimeout);
                    connection.resolve = function () {
                        clearTimeout(connectionTimer);
                        connections.delete(id);
                        resolveConnection();
                    };
                    connection.reject = function (err) {
                        connection.promise.catch(function () { });
                        clearTimeout(connectionTimer);
                        connections.delete(id);
                        rejectConnection(err);
                    };
                });
            };
            connection.reject = function (err) {
                connection.promise.catch(function () { });
                clearTimeout(responseTimer);
                connections.delete(id);
                rejectResponse(err);
            };
            connections.set(id, connection);
        });
        return connection;
    };
    ConnectionsInProgress.prototype.get = function (streamId, id) {
        return this.getByStreamId(streamId).get(id);
    };
    ConnectionsInProgress.prototype.has = function (streamId, id) {
        return this.getByStreamId(streamId).has(id);
    };
    ConnectionsInProgress.prototype.clean = function () {
        this.cleanAll(this.wcStream);
        this.cleanAll(this.sigStream);
    };
    ConnectionsInProgress.prototype.cleanAll = function (connections) {
        connections.forEach(function (c) { return c.reject(new Error(ConnectionError.CLEAN)); });
        connections.clear();
    };
    ConnectionsInProgress.prototype.getByStreamId = function (streamId) {
        return streamId === this.wcStreamId ? this.wcStream : this.sigStream;
    };
    return ConnectionsInProgress;
}());
export { ConnectionsInProgress };
//# sourceMappingURL=ConnectionsInProgress.js.map