var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __values = (this && this.__values) || function (o) {
    var m = typeof Symbol === "function" && o[Symbol.iterator], i = 0;
    if (m) return m.call(o);
    return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
import { ChannelType, MAXIMUM_MISSED_HEARTBEAT } from '../../Channel';
import { isBrowser, log } from '../../misc/util';
import { fullMesh as proto } from '../../proto';
import { ConnectionError } from '../channelBuilder/ConnectionError';
import { Topology, TopologyState } from './Topology';
var REQUEST_MEMBERS_INTERVAL = 20000;
var MAX_ROUTE_DISTANCE = 3;
var HEARTBEAT_INTERVAL = 3000;
var DELAYED_TIMEOUT = 300000;
var FullMesh = (function (_super) {
    __extends(FullMesh, _super);
    function FullMesh(wc) {
        var _this = _super.call(this, wc, FullMesh.SERVICE_ID, proto.Message) || this;
        _this.adjacentMembers = new Map();
        _this.distantMembers = new Map();
        _this.antecedentId = 0;
        _this.delayedMembers = new Set();
        _this.delayedMembersTimers = new Set();
        _this.adjacentBots = new Set();
        _this.heartbeatMsg = _super.prototype.encode.call(_this, { heartbeat: true });
        _this.wcStream.message.subscribe(function (_a) {
            var channel = _a.channel, senderId = _a.senderId, msg = _a.msg;
            return _this.handleServiceMessage(channel, senderId, msg);
        });
        _this.wc.channelBuilder.onConnectionRequest = function (streamId, data) {
            var _a = proto.ConnectionRequest.decode(data), id = _a.id, adjacentIds = _a.adjacentIds;
            if (streamId === _this.wc.STREAM_ID) {
                log.topology("CONNECTION REQUEST from " + id + ", where adjacent members are: ", adjacentIds);
                return _this.createOrUpdateDistantMember(id, adjacentIds);
            }
            else {
                return true;
            }
        };
        _this.wc.channelBuilder.onChannel.subscribe(function (ch) {
            log.topology("New adjacent member: " + ch.id);
            _this.distantMembers.delete(ch.id);
            _this.delayedMembers.delete(ch.id);
            var am = _this.adjacentMembers.get(ch.id);
            if (am) {
                log.topology('Replacing the same channel');
                am.close();
            }
            _this.adjacentMembers.set(ch.id, ch);
            if (ch.url) {
                _this.adjacentBots.add(ch);
            }
            _this.notifyDistantMembers();
            if (!_this.heartbeatInterval) {
                _this.startHeartbeatInterval();
            }
            if (ch.type === ChannelType.JOINING) {
                _super.prototype.setState.call(_this, TopologyState.CONSTRUCTING);
                var members = ch.initData.members;
                _this.connectToMembers(members, ch.id).then(function () {
                    _super.prototype.setState.call(_this, TopologyState.CONSTRUCTED);
                    _this.startMembersCheckIntervals();
                });
            }
            else {
                _this.startMembersCheckIntervals();
            }
            _this.wc.onMemberJoinProxy(ch.id);
            _this.updateAntecedentId();
        });
        if (isBrowser) {
            ;
            window.fullmesh = function () {
                log.topology('Fullmesh info:', {
                    myId: _this.wc.myId,
                    antecedentId: _this.antecedentId,
                    signalingState: _this.wc.signaling.state,
                    webGroupState: _this.wc.state,
                    topologyState: TopologyState[_this.state],
                    adjacentBots: Array.from(_this.adjacentBots).map(function (ch) { return ch.id; }),
                    adjacentMembers: Array.from(_this.adjacentMembers.keys()).toString(),
                    distantMembers: Array.from(_this.distantMembers.keys()).toString(),
                    distantMembersDETAILS: Array.from(_this.distantMembers.keys()).map(function (id) {
                        var adjacentMembers = _this.distantMembers.get(id);
                        return { id: id, adjacentMembers: adjacentMembers ? adjacentMembers.adjacentIds : [] };
                    }),
                });
            };
        }
        return _this;
    }
    FullMesh.prototype.send = function (msg) {
        var _this = this;
        this.adjacentMembers.forEach(function (ch) { return ch.encodeAndSend(msg); });
        this.distantMembers.forEach(function (distantMember, id) {
            _this.sendToDistantPeer(distantMember, Object.assign(msg, { recipientId: id }));
        });
    };
    FullMesh.prototype.sendTo = function (msg) {
        var ch = this.adjacentMembers.get(msg.recipientId);
        if (ch) {
            ch.encodeAndSend(msg);
        }
        else {
            this.sendToDistantPeer(this.distantMembers.get(msg.recipientId), msg);
        }
    };
    FullMesh.prototype.forward = function (msg) {
        this.sendTo(msg);
    };
    FullMesh.prototype.leave = function () {
        if (this.state !== TopologyState.IDLE) {
            log.topology('LEAVE');
            this.clean();
            this.adjacentMembers.forEach(function (ch) { return ch.close(); });
            this.wc.onAdjacentMembersLeaveProxy(Array.from(this.adjacentMembers.keys()));
            this.adjacentMembers.clear();
            _super.prototype.setState.call(this, TopologyState.IDLE);
        }
    };
    FullMesh.prototype.onChannelClose = function (channel) {
        if (this.adjacentMembers.delete(channel.id)) {
            this.adjacentBots.delete(channel);
            this.delayedMembers.delete(channel.id);
            this.wc.onAdjacentMembersLeaveProxy([channel.id]);
            if (this.adjacentMembers.size === 0) {
                this.clean();
            }
            else {
                this.notifyDistantMembers();
                this.updateAntecedentId();
            }
            log.topology("Channel closed with " + channel.id);
        }
    };
    Object.defineProperty(FullMesh.prototype, "neighbors", {
        get: function () {
            return Array.from(this.adjacentMembers.values());
        },
        enumerable: true,
        configurable: true
    });
    FullMesh.prototype.clean = function () {
        log.topology('CLEAN');
        this.wc.onDistantMembersLeaveProxy(Array.from(this.distantMembers.keys()));
        this.distantMembers.clear();
        this.antecedentId = 0;
        clearInterval(this.heartbeatInterval);
        this.heartbeatInterval = undefined;
        clearInterval(this.membersCheckInterval);
        this.membersCheckInterval = undefined;
        this.delayedMembers.clear();
        this.delayedMembersTimers.forEach(function (t) { return clearTimeout(t); });
        this.delayedMembersTimers.clear();
        this.adjacentBots.clear();
    };
    FullMesh.prototype.handleServiceMessage = function (channel, senderId, msg) {
        switch (msg.type) {
            case 'members': {
                this.connectToMembers(msg.members.ids, senderId);
                break;
            }
            case 'adjacentMembers': {
                log.topology("adjacent members received from " + senderId, msg.adjacentMembers.ids);
                var ids = msg.adjacentMembers.ids;
                this.createOrUpdateDistantMember(senderId, ids);
                break;
            }
            case 'heartbeat': {
                var distantMember = this.distantMembers.get(senderId);
                if (distantMember) {
                    distantMember.missedHeartbeat = 0;
                }
                break;
            }
        }
    };
    FullMesh.prototype.connectToMembers = function (ids, adjacentId) {
        var e_1, _a;
        var _this = this;
        var missingIds = ids.filter(function (id) { return !_this.adjacentMembers.has(id) && id !== _this.wc.myId; });
        if (missingIds.length !== 0) {
            log.topology("TRY TO CONNECT to members", missingIds);
            var attempts = [];
            var msg = proto.ConnectionRequest.encode(proto.ConnectionRequest.create({
                id: this.wc.myId,
                adjacentIds: Array.from(this.adjacentMembers.keys()),
            })).finish();
            var _loop_1 = function (id) {
                if (!this_1.distantMembers.has(id)) {
                    this_1.distantMembers.set(id, {
                        adjacentIds: [adjacentId],
                        missedHeartbeat: 0,
                    });
                    log.topology("NEW distant member = " + id + "; BIS", adjacentId);
                    this_1.updateAntecedentId();
                }
                if (!this_1.delayedMembers.has(id)) {
                    attempts[attempts.length] = this_1.wc.channelBuilder
                        .connectOverWebChannel(id, function () {
                        _this.wc.onMemberJoinProxy(id);
                        _this.updateAntecedentId();
                    }, msg)
                        .catch(function (err) {
                        log.topology("FAILED to connect to " + id + ", because: " + err.message);
                        if (err.message === ConnectionError.CONNECTION_TIMEOUT ||
                            err.message === ConnectionError.NEGOTIATION_ERROR ||
                            err.message === ConnectionError.IN_PROGRESS) {
                            _this.delayedMembers.add(id);
                            var timer_1 = setTimeout(function () {
                                _this.delayedMembers.delete(id);
                                _this.delayedMembersTimers.delete(timer_1);
                            }, DELAYED_TIMEOUT);
                            _this.delayedMembersTimers.add(timer_1);
                        }
                    });
                }
            };
            var this_1 = this;
            try {
                for (var missingIds_1 = __values(missingIds), missingIds_1_1 = missingIds_1.next(); !missingIds_1_1.done; missingIds_1_1 = missingIds_1.next()) {
                    var id = missingIds_1_1.value;
                    _loop_1(id);
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (missingIds_1_1 && !missingIds_1_1.done && (_a = missingIds_1.return)) _a.call(missingIds_1);
                }
                finally { if (e_1) throw e_1.error; }
            }
            return Promise.all(attempts);
        }
        return Promise.resolve();
    };
    FullMesh.prototype.notifyDistantMembers = function () {
        var _this = this;
        if (this.distantMembers.size !== 0) {
            var msg_1 = _super.prototype.encode.call(this, {
                adjacentMembers: { ids: Array.from(this.adjacentMembers.keys()) },
            });
            this.distantMembers.forEach(function (dm, id) { return _this.wcStream.send(msg_1, id); });
        }
    };
    FullMesh.prototype.startMembersCheckIntervals = function () {
        var _this = this;
        if (!this.membersCheckInterval) {
            this.membersCheckInterval = setInterval(function () {
                if (_this.antecedentId) {
                    _this.wcStream.send({ members: { ids: _this.wc.members } }, _this.antecedentId);
                }
            }, REQUEST_MEMBERS_INTERVAL);
        }
    };
    FullMesh.prototype.startHeartbeatInterval = function () {
        var _this = this;
        this.heartbeatInterval = setInterval(function () {
            _this.adjacentMembers.forEach(function (ch) { return ch.sendHeartbeat(); });
            _this.distantMembers.forEach(function (peer, id) {
                peer.missedHeartbeat++;
                if (peer.missedHeartbeat >= MAXIMUM_MISSED_HEARTBEAT + 1) {
                    log.topology("Distant peer " + id + " stopped to responding: too many missed heartbeats");
                    if (_this.distantMembers.delete(id)) {
                        _this.delayedMembers.delete(id);
                        _this.wc.onDistantMembersLeaveProxy([id]);
                        _this.updateAntecedentId();
                    }
                }
                _this.wcStream.send(_this.heartbeatMsg, id);
            });
        }, HEARTBEAT_INTERVAL);
    };
    FullMesh.prototype.sendToDistantPeer = function (distantMember, msg) {
        for (var d = 1; d <= MAX_ROUTE_DISTANCE; d++) {
            var ch = this.findRoutedChannel(distantMember, d);
            if (ch) {
                ch.encodeAndSend(msg);
                return;
            }
        }
    };
    FullMesh.prototype.findRoutedChannel = function (distantMember, distance) {
        var e_2, _a, e_3, _b, e_4, _c;
        if (distantMember) {
            try {
                for (var _d = __values(this.adjacentBots), _e = _d.next(); !_e.done; _e = _d.next()) {
                    var ch = _e.value;
                    if (distantMember.adjacentIds.includes(ch.id)) {
                        return ch;
                    }
                }
            }
            catch (e_2_1) { e_2 = { error: e_2_1 }; }
            finally {
                try {
                    if (_e && !_e.done && (_a = _d.return)) _a.call(_d);
                }
                finally { if (e_2) throw e_2.error; }
            }
            try {
                for (var _f = __values(this.adjacentMembers), _g = _f.next(); !_g.done; _g = _f.next()) {
                    var _h = __read(_g.value, 2), neighbourId = _h[0], ch = _h[1];
                    if (distantMember.adjacentIds.includes(neighbourId)) {
                        return ch;
                    }
                }
            }
            catch (e_3_1) { e_3 = { error: e_3_1 }; }
            finally {
                try {
                    if (_g && !_g.done && (_b = _f.return)) _b.call(_f);
                }
                finally { if (e_3) throw e_3.error; }
            }
            if (distance !== 0) {
                try {
                    for (var _j = __values(distantMember.adjacentIds), _k = _j.next(); !_k.done; _k = _j.next()) {
                        var id = _k.value;
                        var ch = this.findRoutedChannel(this.distantMembers.get(id), distance - 1);
                        if (ch) {
                            return ch;
                        }
                    }
                }
                catch (e_4_1) { e_4 = { error: e_4_1 }; }
                finally {
                    try {
                        if (_k && !_k.done && (_c = _j.return)) _c.call(_j);
                    }
                    finally { if (e_4) throw e_4.error; }
                }
            }
        }
        return undefined;
    };
    FullMesh.prototype.createOrUpdateDistantMember = function (id, ids) {
        if (!this.adjacentMembers.has(id)) {
            var distantMember = this.distantMembers.get(id);
            if (distantMember) {
                distantMember.adjacentIds = ids;
                log.topology("UPDATE distant member = " + id, distantMember.adjacentIds);
            }
            else {
                this.distantMembers.set(id, {
                    adjacentIds: ids,
                    missedHeartbeat: 0,
                });
                this.wc.onMemberJoinProxy(id);
                log.topology("NEW distant member = " + id, ids);
                this.updateAntecedentId();
            }
            return true;
        }
        log.topology("ABNORMAL update or create a distant member " + id + ", but he is also an adjacent member");
        return false;
    };
    FullMesh.prototype.updateAntecedentId = function () {
        var e_5, _a;
        if (this.adjacentMembers.size > 0) {
            var maxId = -1;
            var desiredId = -1;
            try {
                for (var _b = __values(this.wc.members), _c = _b.next(); !_c.done; _c = _b.next()) {
                    var id = _c.value;
                    if (id !== this.wc.myId) {
                        if (id < this.wc.myId && id > desiredId) {
                            desiredId = id;
                        }
                        if (maxId < id) {
                            maxId = id;
                        }
                    }
                }
            }
            catch (e_5_1) { e_5 = { error: e_5_1 }; }
            finally {
                try {
                    if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
                }
                finally { if (e_5) throw e_5.error; }
            }
            this.antecedentId = desiredId !== -1 ? desiredId : maxId;
        }
        else {
            this.antecedentId = 0;
        }
        log.topology("UPDATE, antecedentId = " + this.antecedentId + ", members = ", JSON.stringify(this.wc.members));
    };
    FullMesh.SERVICE_ID = 74315;
    return FullMesh;
}(Topology));
export { FullMesh };
//# sourceMappingURL=FullMesh.js.map